#ifndef MSP_GEOMETRY_ANGLE_H_
#define MSP_GEOMETRY_ANGLE_H_

#include <cmath>

namespace Msp {
namespace Geometry {

/**
A planar angle.  Creating an Angle from a raw value or extracting it requires
specifying the units used.  This eliminates a common source of errors.
*/
template<typename T>
class Angle
{
private:
	T value;

	explicit Angle(T v): value(v) { }
public:
	Angle(): value(0) { }
	template<typename U>
	Angle(const Angle<U> &other): value(other.radians()) { }

	static Angle from_degrees(T);
	static Angle from_radians(T);
	static Angle from_turns(T);

	static Angle zero() { return from_radians(0); }
	static Angle right() { return from_turns(0.25); }
	static Angle quarter_turn() { return right(); }
	static Angle straight() { return from_turns(0.5); }
	static Angle half_turn() { return straight(); }
	static Angle full_turn() { return from_turns(1); }

	T degrees() const { return value*57.2957795130823208768; }
	T radians() const { return value; }
	T turns() const { return value/6.28318530717958647692; }

	Angle &operator+=(const Angle &);
	Angle &operator-=(const Angle &);
	Angle &operator*=(T);
	Angle &operator/=(T);

	bool operator==(const Angle &other) const { return value==other.value; }
	bool operator!=(const Angle &other) const { return value!=other.value; }
	bool operator<(const Angle &other) const { return value<other.value; }
	bool operator<=(const Angle &other) const { return value<=other.value; }
	bool operator>(const Angle &other) const { return value>other.value; }
	bool operator>=(const Angle &other) const { return value>=other.value; }

	Angle &wrap_with_base(const Angle &);
	Angle &wrap_positive();
	Angle &wrap_balanced();
};

template<typename T>
inline Angle<T> Angle<T>::from_degrees(T d)
{
	return Angle<T>(d/57.2957795130823208768);
}

template<typename T>
inline Angle<T> Angle<T>::from_radians(T r)
{
	return Angle<T>(r);
}

template<typename T>
inline Angle<T> Angle<T>::from_turns(T t)
{
	return Angle<T>(t*6.28318530717958647692);
}

template<typename T>
inline Angle<T> &Angle<T>::operator+=(const Angle &a)
{
	value += a.value;
	return *this;
}

template<typename T>
inline Angle<T> operator+(const Angle<T> &a1, const Angle<T> &a2)
{
	Angle<T> r(a1);
	return r += a2;
}

template<typename T>
inline Angle<T> &Angle<T>::operator-=(const Angle &a)
{
	value -= a.value;
	return *this;
}

template<typename T>
inline Angle<T> operator-(const Angle<T> &a1, const Angle<T> &a2)
{
	Angle<T> r(a1);
	return r -= a2;
}

template<typename T>
inline Angle<T> operator-(const Angle<T> &a)
{
	return Angle<T>::zero()-a;
}

template<typename T>
inline Angle<T> &Angle<T>::operator*=(T s)
{
	value *= s;
	return *this;
}

template<typename T>
inline Angle<T> operator*(const Angle<T> &a, T s)
{
	Angle<T> r(a);
	return r *= s;
}

template<typename T>
inline Angle<T> operator*(T s, const Angle<T> &a)
{
	return a*s;
}

template<typename T>
inline Angle<T> &Angle<T>::operator/=(T s)
{
	value /= s;
	return *this;
}

template<typename T>
inline Angle<T> operator/(const Angle<T> &a, T s)
{
	Angle<T> r(a);
	return r /= s;
}

template<typename T>
inline Angle<T> operator/(T s, const Angle<T> &a)
{
	return a/s;
}

template<typename T>
inline T operator/(const Angle<T> &a1, const Angle<T> &a2)
{
	return a1.radians()/a2.radians();
}

template<typename T>
inline Angle<T> &Angle<T>::wrap_with_base(const Angle<T> &b)
{
	const float two_pi = 6.28318530717958647692;
	while(value<b.value)
		value += two_pi;
	while(value>=b.value+two_pi)
		value -= two_pi;
	return *this;
}

template<typename T>
inline Angle<T> &Angle<T>::wrap_positive()
{
	return wrap_with_base(zero());
}

template<typename T>
inline Angle<T> &Angle<T>::wrap_balanced()
{
	return wrap_with_base(-half_turn());
}

template<typename T>
inline Angle<T> wrap_with_base(const Angle<T> &a, const Angle<T> &b)
{
	Angle<T> r = a;
	return r.wrap_with_base(b);
}

template<typename T>
inline Angle<T> wrap_positive(const Angle<T> &a)
{
	Angle<T> r = a;
	return r.wrap_positive();
}

template<typename T>
inline Angle<T> wrap_balanced(const Angle<T> &a)
{
	Angle<T> r = a;
	return r.wrap_balanced();
}

template<typename T>
inline Angle<T> abs(const Angle<T> &angle)
{
	using std::abs;
	return Angle<T>::from_radians(abs(angle.radians()));
}

template<typename T>
inline T sin(const Angle<T> &angle)
{
	using std::sin;
	return sin(angle.radians());
}

template<typename T>
inline Angle<T> asin(T s)
{
	using std::asin;
	return Angle<T>::from_radians(asin(s));
}

template<typename T>
inline T cos(const Angle<T> &angle)
{
	using std::cos;
	return cos(angle.radians());
}

template<typename T>
inline Angle<T> acos(T s)
{
	using std::acos;
	return Angle<T>::from_radians(acos(s));
}

template<typename T>
inline T tan(const Angle<T> &angle)
{
	using std::tan;
	return tan(angle.radians());
}

template<typename T>
inline Angle<T> atan(T s)
{
	using std::atan;
	return Angle<T>::from_radians(atan(s));
}

template<typename T>
inline Angle<T> atan2(T y, T x)
{
	using std::atan2;
	return Angle<T>::from_radians(atan2(y, x));
}

} // namespace Geometry
} // namespace Msp

#endif
