#ifndef MSP_GEOMETRY_NEGATION_H_
#define MSP_GEOMETRY_NEGATION_H_

#include "shape.h"

namespace Msp {
namespace Geometry {

/**
Negates a shape.  Not particulary useful on its own, but can be used together
with Intersection to cut holes.
*/
template<typename T, unsigned D>
class Negation: public Shape<T, D>
{
private:
	Shape<T, D> *shape;

public:
	Negation(const Shape<T, D> &);
	Negation(const Negation &);
	Negation &operator=(const Negation &);
	~Negation();

	virtual Negation *clone() const;

	const Shape<T, D> &get_shape() const { return *shape; }

	virtual BoundingBox<T, D> get_axis_aligned_bounding_box() const;
	virtual bool contains(const LinAl::Vector<T, D> &) const;
	virtual unsigned get_max_ray_intersections() const { return shape->get_max_ray_intersections(); }
	virtual unsigned get_intersections(const Ray<T, D> &, SurfacePoint<T, D> *, unsigned) const;
};

template<typename T, unsigned D>
inline Negation<T, D>::Negation(const Shape<T, D> &s):
	shape(s.clone())
{ }

template<typename T, unsigned D>
inline Negation<T, D>::Negation(const Negation<T, D> &other):
	shape(other.shape->clone())
{ }

template<typename T, unsigned D>
inline Negation<T, D> &Negation<T, D>::operator=(const Negation<T, D> &other)
{
	delete shape;
	shape = other.shape->clone();
	return *this;
}

template<typename T, unsigned D>
inline Negation<T, D>::~Negation()
{
	delete shape;
}

template<typename T, unsigned D>
inline Negation<T, D> *Negation<T, D>::clone() const
{
	return new Negation<T, D>(*shape);
}

template<typename T, unsigned D>
inline BoundingBox<T, D> Negation<T, D>::get_axis_aligned_bounding_box() const
{
	return ~shape->get_axis_aligned_bounding_box();
}

template<typename T, unsigned D>
inline bool Negation<T, D>::contains(const LinAl::Vector<T, D> &point) const
{
	return !shape->contains(point);
}

template<typename T, unsigned D>
inline unsigned Negation<T, D>::get_intersections(const Ray<T, D> &ray, SurfacePoint<T, D> *points, unsigned size) const
{
	unsigned count = shape->get_intersections(ray, points, size);
	for(unsigned i=0; i<count; ++i)
	{
		points[i].normal = -points[i].normal;
		points[i].entry = !points[i].entry;
	}
	return count;
}

} // namespace Geometry
} // namespace Msp

#endif
